@@ -4,19 +4,247 @@
 # -----------------------------
 
 
+import copy 
 from urllib.parse import urlparse
 from typing import Any, List, Union
-from . import adapters, lib, cursor, jsonquery, types, operators
+from . import adapters, dictquery, lib, cursor, types, operators, exceptions
 
 
 SQL_DATETIME_FORMAT = "YYYY-MM-DD HH:mm:ss"
+ADD_MANY_CHUNK_SIZE = 1000
+
+class __Item(dict):
+    NAMESPACE = None
+
+    def _make_path(self, path):
+        # if self.NAMESPACE:
+        #     return "%s.%s" % (self.NAMESPACE, path)
+        return path
+
+    def _update(self, data):
+        raise NotImplementedError()
+
+    def get(self, path: str, default: Any = None) -> Any:
+        """
+        GET: Return a property by key/DotNotation
+
+        ie: 
+            #get("key.deep1.deep2.deep3")
+
+        Args:
+            path:str - the dotnotation path
+            default:Any - default value 
+
+        Returns:
+            Any
+        """
+        path = self._make_path(path)
+        return lib.dict_get(obj=dict(self), path=path, default=default)
+
+    def set(self, path: str, value: Any):
+        """
+        SET: Set a property by key/DotNotation
+
+        Args:
+            path:str - the dotnotation path
+            value:Any - The value
+
+        Returns:
+            Void
+        """
+
+        path = self._make_path(path)
+        self._update({path: value})
+
+    def len(self, path: str):
+        """
+        Get the length of the items in a str/list/dict
+        Args:
+            path:str - the dotnotation path
+        Returns:
+            data that was removed
+        """
+        path = self._make_path(path)
+        v = self.get(path)
+        return len(v) if v else 0
+
+    def incr(self, path: str, incr=1):
+        """
+        INCR: increment a value by 1
+        Args
+            path:str - path
+            incr:1 - value to inc by
+        Returns:    
+            int - the value that was incremented
+        """
+        op = "%s:$incr" % self._make_path(path)
+        oplog = self._update({op: incr})
+        return oplog.get(op)
+
+    def decr(self, path: str, decr=1):
+        """
+        DECR: decrement a value by 1
+        Args
+            path:str - path
+            decr:1 - value to dec by
+        Returns:    
+            int - the value that was decremented
+        """
+        op = "%s:$decr" % self._make_path(path)
+
+        oplog = self._update({op: decr})
+        return oplog.get(op)
+
+    def unset(self, path: str):
+        """ 
+        UNSET: Remove a property by key/DotNotation and return the value
+
+        Args:
+            path:str
+
+        Returns:
+            Any: the value that was removed
+        """
+        path = self._make_path(path)
+        self._update({"%s:$unset" % path: True})
+
+    def xadd(self, path: str, values):
+        """
+        LADD: Add *values if they don't exist yet
+
+        Args:
+            path:str - the dotnotation path
+            *values: set of items
+        Returns:
+            list: updated data
+        """
+        op = "%s:$xadd" % self._make_path(path)
+        self._update({op: values})
+
+    def xadd_many(self, path: str, *values: List[Any]):
+        """
+        LADD: Add *values if they don't exist yet
+
+        Args:
+            path:str - the dotnotation path
+            *values: set of items
+        Returns:
+            list: updated data
+        """
+        op = "%s:$xadd_many" % self._make_path(path)
+        self._update({op: values})
+
+    def xrem(self, path: str, values):
+        """
+        LREM: Remove items from a list
+
+        Args:
+            path:str - the dotnotation path
+            *values: set of items
+        Returns:
+            list: updated data
+        """
+        op = "%s:$xrem" % self._make_path(path)
+        oplog = self._update({op: values})
+        return oplog.get(op)
+
+    def xrem_many(self, path: str, *values: List[Any]):
+        """
+        LREM: Remove items from a list
+
+        Args:
+            path:str - the dotnotation path
+            *values: set of items
+        Returns:
+            list: updated data
+        """
+        op = "%s:$xrem_many" % self._make_path(path)
+        oplog = self._update({op: values})
+        return oplog.get(op)
+
+    def xpush(self, path: str, values: Any):
+        """
+        LPUSH: push item to the right of list. 
+
+        Args:
+            path:str - the dotnotation path
+            *values: set of items
+        Returns:
+            list: updated data
+        """
+        op = "%s:$xpush" % self._make_path(path)
+        self._update({op: values})
+
+    def xpush_many(self, path: str, *values: List[Any]):
+        """
+        LPUSH: push item to the right of list. 
+
+        Args:
+            path:str - the dotnotation path
+            *values: set of items
+        Returns:
+            list: updated data
+        """
+        op = "%s:$xpush_many" % self._make_path(path)
+        self._update({op: values})
+
+    def xpushl(self, path: str, values: Any):
+        """
+        LPUSH: push item to the right of list. 
+
+        Args:
+            path:str - the dotnotation path
+            *values: set of items
+        Returns:
+            list: updated data
+        """
+        op = "%s:$xpushl" % self._make_path(path)
+        self._update({op: values})
+
+    def xpushl_many(self, path: str, *values: List[Any]):
+        """
+        LPUSH: push item to the right of list. 
+
+        Args:
+            path:str - the dotnotation path
+            *values: set of items
+        Returns:
+            list: updated data
+        """
+        op = "%s:$xpush_many" % self._make_path(path)
+        self._update({op: values})
+
+    def xpop(self, path: str):
+        """
+        Remove value at the end an array/list
+        Args:
+            path:str - the dotnotation path
+        Returns:
+            data that was removed
+
+        """
+        op = "%s:$xpop" % self._make_path(path)
+        oplog = self._update({op: True})
+        return oplog.get(op)
+
+    def xpopl(self, path: str):
+        """
+        Remove value at the beginning an array/list
+        Args:
+            path:str - the dotnotation path
+        Returns:
+            data that was removed        
+        """
+        op = "%s:$xpopl" % self._make_path(path)
+        oplog = self._update({op: True})
+        return oplog.get(op)
 
 
 class Database(object):
     """
     :: Database
 
-    Class to create a connection to an adapter and select a container
+    Class to create a connection to an adapter and select a collection
     """
 
     def __init__(self, dsn_adapter: Union[str, adapters.BaseAdapter]):
@@ -42,297 +270,115 @@ class Database(object):
             elif _.scheme in ["mysql", "mariadb"]:
                 self.db = adapters.MySQLAdapter(dsn_adapter)
 
-    def container(self, name: str, columns: list = None):
+
+    def collection(self, name: str, columns: list = None):
         """
-        Select a container
+        Select a collection
 
         Args:
-            name: str - container name
+            name: str - collection name
             columns: list - list of columns and indexes to add
 
         Returns:
-            Container
+            Collection
         """
-        container = Container(self.db, name)
+        collection = Collection(self.db, name)
         if columns:
-            container.add_columns(columns)
-        return container
+            collection.add_columns(columns)
+        return collection
 
-    def drop_container(self, name):
+    def drop_collection(self, name):
         """
-        Drop/Delete a table/container
+        Drop/Delete a table/collection
 
         Returns:
             None
         """
-        self.db.drop_container(name)
+        self.db.drop_collection(name)
 
     @property
-    def containers(self) -> list:
+    def collections(self) -> list:
         """
-        List containers in the database
+        List collections in the database
 
         Returns: 
           list
         """
-        return self.db.get_containers()
+        return self.db.get_collections()
 
     @staticmethod
     def utcnow():
         return lib.get_timestamp()
 
 
-class Item(dict):
+class CollectionItem(__Item):
     """
-    Item/ActiveRecord
+    CollectionItem/ActiveRecord
 
     Every row is a document in Kongodb
     """
+    SUBCOLLECTION_NAMESPACE = "/subcollection/%s"
 
-    def __init__(self, container, row: dict):
-        self.container = container
+    def __init__(self, collection, row: dict):
+        self._collection = collection
         self._load(row)
 
-
-    def get(self, path: str, default: Any = None) -> Any:
+    def update(self, data: dict):
         """
-        GET: Return a property by key/DotNotation
-
-        ie: 
-            #get("key.deep1.deep2.deep3")
-
-        Args:
-            path:str - the dotnotation path
-            default:Any - default value 
+        UPDATE: Update the active CollectionItem
 
         Returns:
-            Any
-        """
-        return lib.dict_get(obj=dict(self), path=path, default=default)
-
-    def set(self, path: str, value: Any):
-        """
-        SET: Set a property by key/DotNotation
-
-        Args:
-            path:str - the dotnotation path
-            value:Any - The value
-
-        Returns:
-            Void
-        """
-        self._update({path : value })
-
-    def len(self, path: str):
-        """
-        Get the length of the items in a str/list/dict
-        Args:
-            path:str - the dotnotation path
-        Returns:
-            data that was removed
-        """
-        v = self.get(path)
-        return len(v) if v else 0
-
-    def unset(self, path: str):
-        """ 
-        UNSET: Remove a property by key/DotNotation and return the value
-
-        Args:
-            path:str
-
-        Returns:
-            Any: the value that was removed
-        """
-        self._update({ path + ":$unset" % path: True })
-        
-    def update(self, data:dict):
-        """
-        UPDATE: Update the active Item
-
-        Returns:
-            Item
+            CollectionItem
         """
         self._update(data)
 
     def delete(self):
         """
-        DELETE: Delete the Doument from the container
+        DELETE: Delete the Doument from the collection
 
         Returns:
             None
         """
-        self.container.delete_id(self._id)
+        self._collection.delete_id(self._id)
         self._clear_self()
 
-    # ---
+    def subcollection(self, name: str, constraints: list = None):
+        collection_info = _CollectionInfo__(self._collection, name)
+        _constraints = collection_info.get_subcollection_constraints(name) or []
+        if constraints and set(constraints) != set(_constraints):
+            collection_info.set_subcollection_constraints(name, constraints)
+            _constraints = constraints
+        return SubCollection(item=self, name=name, constraints=_constraints)
 
-    def incr(self, path: str, incr=1):
-        """
-        INCR: increment a value by 1
-        Args
-            path:str - path
-            incr:1 - value to inc by
-        Returns:    
-            int - the value that was incremented
-        """
-        op = "%s:$incr" % path
-        oplog = self._update({op: incr })
-        return oplog.get(op)
+    @property
+    def subcollections(self):
+        return list(self.get("/collections").keys()) or []
 
-    def decr(self, path: str, decr=1):
-        """
-        DECR: decrement a value by 1
-        Args
-            path:str - path
-            decr:1 - value to dec by
-        Returns:    
-            int - the value that was decremented
-        """
-        op = "%s:$decr" % path
-        oplog = self._update({op: decr })
-        return oplog.get(op)
+    def drop_subcollection(self, name: str):
+        try:
+            self.unset("/collections.%s" % name)
+        except KeyError as _:
+            pass
+        return True
 
-    def xadd(self, path: str, values):
-        """
-        LADD: Add *values if they don't exist yet
-
-        Args:
-            path:str - the dotnotation path
-            *values: set of items
-        Returns:
-            list: updated data
-        """
-        op = "%s:$xadd" % path
-        self._update({op: values })
-
-    def xadd_multi(self, path: str, *values: List[Any]):
-        """
-        LADD: Add *values if they don't exist yet
-
-        Args:
-            path:str - the dotnotation path
-            *values: set of items
-        Returns:
-            list: updated data
-        """
-        op = "%s:$xadd_multi" % path
-        self._update({op: values })
-
-    def xrem(self, path: str, values):
-        """
-        LREM: Remove items from a list
-
-        Args:
-            path:str - the dotnotation path
-            *values: set of items
-        Returns:
-            list: updated data
-        """
-        op = "%s:$xrem" % path
-        oplog = self._update({op: values })
-        return oplog.get(op)
-
-    def xrem_multi(self, path: str, *values: List[Any]):
-        """
-        LREM: Remove items from a list
-
-        Args:
-            path:str - the dotnotation path
-            *values: set of items
-        Returns:
-            list: updated data
-        """
-        op = "%s:$xrem_multi" % path
-        oplog = self._update({op: values })
-        return oplog.get(op)
-
-    def xpush(self, path: str, values:Any):
-        """
-        LPUSH: push item to the right of list. 
-
-        Args:
-            path:str - the dotnotation path
-            *values: set of items
-        Returns:
-            list: updated data
-        """
-        op = "%s:$xpush" % path
-        self._update({op: values })
-
-    def xpush_multi(self, path: str, *values: List[Any]):
-        """
-        LPUSH: push item to the right of list. 
-
-        Args:
-            path:str - the dotnotation path
-            *values: set of items
-        Returns:
-            list: updated data
-        """
-        op = "%s:$xpush_multi" % path
-        self._update({op: values })
-
-    def lpushl(self, path: str, *values: List[Any]):
-        """
-        LPUSHL: push item to the left
-
-        Args:
-            path:str - the dotnotation path
-            *values: set of items
-        Returns:
-            list: updated data
-        """
-        v = self._req_list_prop(path)
-        v2 = list(values)
-        v2.extend(v)
-        self.set(path=path, value=v2)
-        return self.get(path)
-
-    def lpop(self, path: str):
-        """
-        Remove value at the end an array/list
-        Args:
-            path:str - the dotnotation path
-        Returns:
-            data that was removed
-
-        """
-        op = "%s:$xpop" % path
-        oplog = self._update({op: True })
-        return oplog.get(op)
-
-    def lpopl(self, path: str):
-        """
-        Remove value at the beginning an array/list
-        Args:
-            path:str - the dotnotation path
-        Returns:
-            data that was removed        
-        """
-        op = "%s:$xpopl" % path
-        oplog = self._update({op: True })
-        return oplog.get(op)
-
-
-    # --- misc
     def save(self):
         """
         To commit the data when it's mutated outside.
-            doc = Item()
+            doc = CollectionItem()
             doc["xone"][1] = True
             doc.save()
         """
         data = dict(self)
-        self.update(data)
+        self._update(data)
 
-    def _update(self, ops:dict):
+    def _update(self, ops: dict):
         """
         Return oplog
         """
         data = dict(self)
         doc, oplog = operators.update_operations(data, ops)
-        row = self.container.update_id(_id=self._id, doc=doc, _as_document=False)
+        row = self._collection.update_id(
+            _id=self._id, doc=doc, replace=True, _as_document=False)
         self._load(row)
         return oplog
 
@@ -355,9 +401,9 @@ class Item(dict):
                 del self[_]
 
 
-class Container(object):
+class Collection(object):
     """
-    Container/Table
+    Collection/Table
 
     """
 
@@ -368,7 +414,7 @@ class Container(object):
     def __init__(self, conn: adapters.BaseAdapter, name):
         self.name = name
         self.db = conn
-        self.db.create_container(self.name)
+        self.db.create_collection(self.name)
 
     def __len__(self):
         return self.size
@@ -398,7 +444,7 @@ class Container(object):
     @property
     def size(self) -> int:
         """
-        Get the total entries in the container
+        Get the total entries in the collection
 
         Returns:
             int
@@ -407,27 +453,27 @@ class Container(object):
 
     # ---- methods ----
 
-    def get(self, _id: str) -> Item:
+    def get(self, _id: str) -> CollectionItem:
         """
         Get a document by _id
         Alias to find_one(_id)
 
         Returns:
-            Item
+            CollectionItem
         """
         return self.find_one(_id=_id)
 
-    def find_one(self, *a, **kw) -> Item:
+    def find_one(self, *a, **kw) -> CollectionItem:
         """
         Retrieve 1 document by _id, or other filters
 
         Args:
           _id:str - the document id
-          _as_document:bool - when True return Item
+          _as_document:bool - when True return CollectionItem
           **kw other query
 
         Returns:
-          Item
+          CollectionItem
 
         Examples:
             #.find_one()
@@ -443,23 +489,27 @@ class Container(object):
             _as_document = kw.pop("_as_document")
 
         # expecting the first arg to be _id
-        filters = {"_id": a[0]} if a else kw
-
+        filters = kw or {} 
+        if a and a[0]:
+            if isinstance(a[0], dict):
+                filters = a[0]
+            else: 
+                filters = {"_id": a[0]}
         r = self.find(filters=filters, limit=1, _as_document=_as_document)
         return r[0] if len(r) else None
 
-    def find_all(self, *a, **kw) -> List[Item]:
+    def find_all(self, *a, **kw) -> List[CollectionItem]:
         """
         Retrieve all docuemts based on criteria return in a list
 
         Returns:
-            List[Item]
+            List[CollectionItem]
         """
         return list(self.find(*a, **kw))
 
-    def find(self, filters: dict = {}, sorts:dict = {}, limit: int = 10, skip: int = 0, _as_document=True) -> cursor.Cursor:
+    def find(self, filters: dict = {}, sorts: dict = {}, limit: int = 10, skip: int = 0, _as_document=True) -> cursor.Cursor:
         """
-        To fetch data from the container
+        To fetch data from the collection
 
         Smart Query
           Allow to use primary indexes from sqlite 
@@ -480,7 +530,7 @@ class Container(object):
         # If there is more search properties, take it to the json
         xparams = []
         xquery = []
-        
+
         smart_filters = _parse_smart_filtering(
             filters, indexed_columns=self.columns)
 
@@ -507,13 +557,13 @@ class Container(object):
             for chunked in self.db.fetchmany(query, xparams, chunk):
                 if chunked:
                     rows = [_parse_row(row) for row in chunked]
-                    for r in jsonquery.execute(rows, smart_filters["JSON_FILTERS"]):
+                    for r in _query_dict(rows, smart_filters["JSON_FILTERS"]):
                         data.append(r)
                 else:
                     break
             if data:
                 if _as_document:
-                    data = [Item(self, d) for d in data]
+                    data = [CollectionItem(self, d) for d in data]
             sorts = _parse_sort_dict(sorts, False)
             return cursor.Cursor(data, sort=sorts, limit=limit, skip=skip)
 
@@ -534,56 +584,93 @@ class Container(object):
 
             res = self.db.fetchall(query, xparams)
             if _as_document:
-                data = [Item(self, row) for row in res]
+                data = [CollectionItem(self, row) for row in res]
             else:
                 data = list(res)
             return cursor.Cursor(data)
 
-    def add(self, doc: dict, _as_document: bool = True) -> Item:
+    def add(self, data: dict, _as_document: bool = True) -> CollectionItem:
         """
-        Add a new document in container
+        Add a new document in collection
 
         use Smart Insert, by checking if a value in the doc in is a column.
 
         Args:
-          doc:dict - Data to be inserted
+          data:dict - Data to be inserted
 
         Returns:
-            Item
+            CollectionItem
         """
-        if not isinstance(doc, dict):
+        if not isinstance(data, dict):
             raise TypeError('Invalid data type. Must be a dict')
 
-        _id = lib.gen_id()
-        ts = lib.get_timestamp()
-        f_ts = ts.format(SQL_DATETIME_FORMAT)
-        doc.update({
-            "_id": _id,
-            "_created_at": ts,
-            "_modified_at": ts,
-        })
+        stmt, xparams = self._prepare_add(data=[data])
+        self.db.execute(stmt, xparams[0])
+        return self.find_one(_id=xparams[0][0], _as_document=_as_document)
 
-        xcolumns = ["_id", "_json", "_created_at", "_modified_at"]
-        xparams = [_id, lib.json_dumps(doc), f_ts, f_ts]
-        q = "INSERT INTO %s " % self.name
+    def add_many(self, data:List[dict]) -> bool:
+        """
+        Add many document at once.
 
-        # indexed data
-        # some data can't be overwritten
-        for col in self.columns:
-            if col in doc and col not in xcolumns:
-                xcolumns.append(col)
-                xparams.append(doc[col])
+        Args:
+            data:list[dict]
+        """
 
-        q += " ( %s ) VALUES ( %s ) " % (",".join(xcolumns),
-                                         ",".join(["?" for _ in xparams]))
+        if not isinstance(data, list):
+            raise TypeError('Invalid data type. Must be a list')
 
-        self.db.execute(q, xparams)
-        return self.find_one(_id=_id, _as_document=_as_document)
+        stmt, xparams = self._prepare_add(data=[data])
+        for xdata in lib.chunk_list(xparams, ADD_MANY_CHUNK_SIZE):
+            self.db.execute(stmt, xdata, many=True)
+        return True
 
-    def update(self, filters): raise NotImplementedError()
+    def _prepare_add(self, data:list) -> tuple:
+        """
+        Prepare ADD/INSERT statement
+
+        Returns 
+            tuple(statement:str, data:List[list])
+        """
+
+        if not isinstance(data, list):
+            raise TypeError('Invalid data type. Must be a list')
+
+        data = copy.deepcopy(data)
+
+        table_columns = self.columns
+        _columns = None
+        xparams = []
+        for d in data:
+            _id = lib.gen_id()
+            ts = lib.get_timestamp()
+            f_ts = ts.format(SQL_DATETIME_FORMAT)
+            datax, _ = operators.update_operations({}, d)
+            datax.update({
+                "_id": _id,
+                "_created_at": ts,
+                "_modified_at": ts,
+            })
+
+            _columns = ["_id", "_json", "_created_at", "_modified_at"]
+            _params = [_id, lib.json_dumps(datax), f_ts, f_ts]
+        
+            # indexed data
+            # some data can't be overwritten
+            for col in table_columns:
+                if col in datax and col not in _columns:
+                    _columns.append(col)
+                    _params.append(datax[col])
+
+            xparams.append(_params)
+
+        stmt = "INSERT INTO %s " % self.name
+        stmt += " ( %s ) VALUES ( %s ) " % (",".join(_columns), ",".join(["?" for _ in xparams[0]]))
+        return (stmt, xparams)
+
+    def update(self, filters, data:dict): raise NotImplementedError()
     def delete(self, filters): raise NotImplementedError()
 
-    def update_id(self, _id: str, doc: dict = {}, replace: bool = False, _as_document=True) -> Item:
+    def update_id(self, _id: str, doc: dict = {}, replace: bool = False, _as_document=True) -> CollectionItem:
         """
         To update a document
 
@@ -594,34 +681,35 @@ class Container(object):
                   When True, it will save it as is. 
 
         Returns:
-            Item
+            CollectionItem
         """
-        rdoc = self.find_one(_id=_id, _as_document=False)
-        if rdoc:
-            ts = lib.get_timestamp()
-            ts_f = ts.format(SQL_DATETIME_FORMAT)
-            _doc = doc if replace else lib.dict_merge(
-                lib.json_loads(rdoc["_json"]), doc)
-            _doc["_modified_at"] = ts
-            _restricted_columns = self.DEFAULT_COLUMNS[:]
+        ts = lib.get_timestamp()
+        ts_f = ts.format(SQL_DATETIME_FORMAT)
 
-            xcolumns = ["_json", "_modified_at"]
-            xparams = [lib.json_dumps(_doc), ts_f]
+        if replace:
+            _doc = doc
+        else:
+            rdoc = self.find_one(_id=_id, _as_document=False)
+            _doc = lib.dict_merge(lib.json_loads(rdoc["_json"]), doc)
 
-            q = "UPDATE %s SET " % self.name
+        _doc["_modified_at"] = ts
+        _restricted_columns = self.DEFAULT_COLUMNS[:]
+        xcolumns = ["_json", "_modified_at"]
+        xparams = [lib.json_dumps(_doc), ts_f]
 
-            # indexed data
-            # some data can't be overriden
-            for col in self.columns:
-                if col in _doc and col not in _restricted_columns:
-                    xcolumns.append(col)
-                    xparams.append(_doc[col])
-            q += ",".join(["%s = ?" % _ for _ in xcolumns])
-            q += " WHERE _id=?"
-            xparams.append(_id)
-            self.db.execute(q, xparams)
-            return self.find_one(_id=_id, _as_document=_as_document)
-        return None
+        q = "UPDATE %s SET " % self.name
+
+        # indexed data
+        # some data can't be overriden
+        for col in self.columns:
+            if col in _doc and col not in _restricted_columns:
+                xcolumns.append(col)
+                xparams.append(_doc[col])
+        q += ",".join(["%s = ?" % _ for _ in xcolumns])
+        q += " WHERE _id=?"
+        xparams.append(_id)
+        self.db.execute(q, xparams)
+        return self.find_one(_id=_id, _as_document=_as_document)
 
     def delete_id(self, _id: str) -> bool:
         """
@@ -663,11 +751,144 @@ class Container(object):
         self.add_columns(columns=columns, enforce_index=True)
 
     # TODO:
-    # def insert_many(self, data:List[dict])
     # def update_many(self, data:dict, filters:dict)
     # def delete_many()
 
 
+class SubCollection(object):
+    _data = []
+    _constraints = []
+
+    def __init__(self, item: CollectionItem, name: str, constraints:list=None):
+        self._item = item
+        self._name = name
+        self._collection_name = "/collections.%s" % name
+        self._constraints = constraints
+        self._load()
+
+    def _load(self):
+        self._data = self._item.get(self._collection_name) or []
+
+    def _commit(self):
+        self._item.set(self._collection_name, self._data)
+
+    def _save(self, _id, data):
+        _data = self._normalize_data()
+        _data[_id] = data
+        self._data = self._denormalize_data(_data)
+        self._commit()        
+
+    def _normalize_data(self) -> dict:
+        return { d.get("_id"): d for d in self._data}
+
+    def _denormalize_data(self, data:dict) -> list:
+        return list(data.values())
+
+    def __len__(self):
+        return len(self._data)
+
+    def drop_subcollection(self):
+        self._item.unset(self._collection_name)
+
+    def add(self, data: dict):
+        data, _ = operators.update_operations({}, data.copy())
+
+        if self._constraints:
+            for c in  self._constraints:
+                if c in data:
+                    if self.find_one({c: data[c]}):
+                        raise exceptions.ConstraintError("Key: %s" % c)
+
+
+        _id = lib.gen_id()
+        ts = lib.get_timestamp()
+        doc = {
+            **data,
+            "_id": _id,
+            "_created_at": ts,
+            "_modified_at": ts
+        }
+        self._data.append(doc)
+        self._commit()
+        return SubCollectionItem(self, doc)
+
+    def update(self, filters, data_ops: dict):
+        _data = self._normalize_data()
+        for item in self.find(filters):
+            ts = lib.get_timestamp()
+            _id = item.get("_id")
+            _default = {  # ensuring we do some data can't be overwritten
+                "_id": _id,
+                "_created_at": ts,
+                "_modified_at": ts
+            }
+            upd, _ = operators.update_operations(item, data_ops)
+            _data[_id] = {**upd, **_default}
+        self._data = self._denormalize_data(_data)
+        self._commit()
+
+    def delete(self, filters: dict):
+        _data = self._normalize_data()
+        for item in self.find(filters):
+            del _data[item.get("_id")]
+        self._data = self._denormalize_data(_data)
+        self._commit()
+
+    def find(self, filters: dict = {}, sorts: dict = {}, limit: int = 10, skip: int = 0) -> cursor.Cursor:
+        """
+        Args:
+            filters:
+            sorts:
+            limit:
+            skip:
+        """
+        data = _query_dict(data=self._data, filters=filters)
+        data = [SubCollectionItem(self, d) for d in data]
+        sorts = _parse_sort_dict(sorts, False)
+        return cursor.Cursor(data, sort=sorts, limit=limit, skip=skip)
+
+
+class SubCollectionItem(__Item):
+
+    def __init__(self, subCollection: SubCollection, data):
+        self.subCollection = subCollection
+        self.NAMESPACE = self.subCollection._collection_name
+        self._load(data)
+
+    def _update(self, ops):
+        data = dict(self)
+        doc, oplog = operators.update_operations(data, ops)
+        self.subCollection._save(self._id, doc)
+        self._load(doc)
+        return oplog
+
+    def _load(self, data):
+        self._id = data.get("_id")
+        super().__init__(data)
+
+class _CollectionInfo__(object):
+    COLLECTION_NAME = "__collections_info__"
+
+    def __init__(self, collection:Collection, collection_name:str):
+        coll = Collection(collection.db, self.COLLECTION_NAME)
+        coll.add_columns([types.StringType("collection_name", unique=True)])
+
+        self.item = coll.find_one({"collection_name": collection_name})
+        if not self.item:
+            self.item = coll.add({"collection_name": collection_name})
+
+    def get_subcollection_constraints(self, subcollection_name):
+        path = "constraints.%s" % subcollection_name
+        return self.item.get(path) or []
+
+    def set_subcollection_constraints(self, subcollection_name, constraints:list):
+        path = "constraints.%s" % subcollection_name
+        if constraints and isinstance(constraints, list):
+            return self.item.set(path, constraints)
+
+
+# ---------------------------------------------------------------------------------------
+# ---------------------------------------------------------------------------------------
 
 def _parse_smart_filtering(filters: dict, indexed_columns: list = []) -> dict:
     """
@@ -694,14 +915,14 @@ def _parse_smart_filtering(filters: dict, indexed_columns: list = []) -> dict:
     sql_filters = []
     json_filters = {}
     for k, v in filters.items():
-        if jsonquery.FILTER_OPERATOR in k:
-            f, op = k.split(jsonquery.FILTER_OPERATOR)
-            if f in indexed_columns and op in jsonquery.SQL_OPERATORS:
-                sql_filters.append((f, jsonquery.SQL_OPERATORS[op], v))
+        if dictquery.FILTER_OPERATOR in k:
+            f, op = k.split(dictquery.FILTER_OPERATOR)
+            if f in indexed_columns and op in dictquery.SQL_OPERATORS:
+                sql_filters.append((f, dictquery.SQL_OPERATORS[op], v))
                 continue
         else:
             if k in indexed_columns:
-                sql_filters.append((k, jsonquery.SQL_OPERATORS["eq"], v))
+                sql_filters.append((k, dictquery.SQL_OPERATORS["eq"], v))
                 continue
         json_filters[k] = v
 
@@ -728,15 +949,22 @@ def _parse_row(row: dict) -> dict:
         **_json
     }
 
+
 def _ascdesc(v, as_str=True):
     if as_str:
         if isinstance(v, int):
-            return "DESC" if v == -1 else "ASC" 
+            return "DESC" if v == -1 else "ASC"
     else:
         if isinstance(v, str):
             return -1 if v.upper() == "DESC" else 1
     return v
 
-def _parse_sort_dict(sorts:dict, as_str=True):
-    return [(k, _ascdesc(v, as_str)) for k,v in sorts.items()]
 
+def _parse_sort_dict(sorts: dict, as_str=True):
+    return [(k, _ascdesc(v, as_str)) for k, v in sorts.items()]
+
+
+def _query_dict(data:list, filters:dict) -> List[dict]:
+    if not filters:
+        return data
+    return dictquery.DictQuery(data).execute(filters)
\ No newline at end of file
